@@ -165,6 +165,7 @@ using SchemeVec = std::vector<ComputeScheme>;
 static SchemeVec EmptySchemeVec = {};
 
 class TreeNode;
+class LeafNode;
 
 // The mininal tree node data needed to decide the scheme
 struct NodeMetaData
@@ -178,12 +179,13 @@ struct NodeMetaData
     size_t                  iDist = 0, oDist = 0;
     size_t                  iDistBlue = 0, oDistBlue = 0;
     size_t                  iOffset = 0, oOffset = 0;
-    int                     direction    = -1;
-    rocfft_result_placement placement    = rocfft_placement_inplace;
-    rocfft_precision        precision    = rocfft_precision_single;
-    rocfft_array_type       inArrayType  = rocfft_array_type_unset;
-    rocfft_array_type       outArrayType = rocfft_array_type_unset;
-    hipDeviceProp_t         deviceProp   = {};
+    bool                    applyPartialPass = false;
+    int                     direction        = -1;
+    rocfft_result_placement placement        = rocfft_placement_inplace;
+    rocfft_precision        precision        = rocfft_precision_single;
+    rocfft_array_type       inArrayType      = rocfft_array_type_unset;
+    rocfft_array_type       outArrayType     = rocfft_array_type_unset;
+    hipDeviceProp_t         deviceProp       = {};
     bool                    rootIsC2C;
 
     explicit NodeMetaData(TreeNode* refNode);
@@ -374,6 +376,9 @@ public:
     size_t lengthBlue  = 0;
     size_t lengthBlueN = 0;
 
+    // enables partial pass for this node
+    bool applyPartialPass = false;
+
     //
     BluesteinType     typeBlue   = BluesteinType::BT_NONE;
     BluesteinFuseType fuseBlue   = BluesteinFuseType::BFT_NONE;
@@ -385,6 +390,8 @@ public:
     size_t           twiddles_size       = 0;
     void*            twiddles_large      = nullptr;
     size_t           twiddles_large_size = 0;
+    void*            twiddles_pp         = nullptr;
+    size_t           twiddles_pp_size    = 0;
     void*            chirp               = nullptr;
     size_t           chirp_size          = 0;
     gpubuf_t<size_t> devKernArg;
@@ -445,6 +452,10 @@ public:
     bool isOutArrayTypeAllowed(rocfft_array_type) const;
     bool isRootNode() const;
     bool isLeafNode() const;
+    // Get the current node downcasted to a LeafNode.  Throws bad_cast
+    // if this node isn't actually a leaf.
+    LeafNode&       getLeafNode();
+    const LeafNode& getLeafNode() const;
 
     // whether or not the input/output access pattern may benefit from padding
     virtual bool PaddingBenefitsInput()
@@ -641,6 +652,7 @@ public:
     bool                twd_no_radices   = false;
     bool                twd_attach_halfN = false;
     std::vector<size_t> kernelFactors    = {};
+    std::vector<size_t> kernelFactorsPP  = {}; // factors for off-direction partial pass(es)
     size_t              bwd              = 1; // bwd, wgs, lds are for grid param lds_bytes
     size_t              wgs              = 0;
     size_t              lds              = 0;
@@ -670,6 +682,7 @@ public:
     void         SetupGridParamAndFuncPtr(DevFnCall& fnPtr, GridParam& gp) override;
     FMKey        GetKernelKey() const override;
     virtual void GetKernelFactors();
+    virtual void GetKernelPartialPassFactors();
 };
 
 /*****************************************************
@@ -962,24 +975,28 @@ struct MultiPlanItem
         return 0;
     }
 
-    // print a description of this item to the plan log
+    // Print a description of this item to the plan log
     virtual void Print(rocfft_ostream& os, const int indent) const = 0;
 
     // utility function to print a buffer enum with a description of
     // the pointer and an offset
     static std::string PrintBufferPtrOffset(const BufferPtr& ptr, size_t offset);
 
-    // check if this item writes to the specified BufferPtr
+    // Check if this item writes to the specified BufferPtr
     virtual bool WritesToBuffer(const BufferPtr& ptr) const = 0;
 
-    // check if this the specified rank will execute this item
+    // Check if the specified rank will execute this item
     virtual bool ExecutesOnRank(int rank) const = 0;
+    // Check if local rank will execute this item
+    bool ExecutesOnLocalRank() const
+    {
+        return ExecutesOnRank(local_comm_rank);
+    };
 
-    // high-level description of what this item is doing, displayed
-    // when logging plan graph
+    // High-level description of what this item is doing, displayed when logging plan graph
     std::string description;
-    // group to assign this item to (letters, numbers, underscores).
-    // items in the same group are drawn together in the graph
+    // Group to assign this item to (letters, numbers, underscores).
+    // Items in the same group are drawn together in the graph
     std::string group;
 
     // Compute a communication tag for an operation, for an item with
@@ -988,6 +1005,9 @@ struct MultiPlanItem
     // single operation in it.  This function produces unique tags for
     // items with multiple operations.
     static int GetOperationCommTag(size_t multiPlanIdx, size_t opIdx);
+
+    // This process's rank relative to the plan communicator.
+    int local_comm_rank;
 };
 
 // Communication operations
@@ -1171,6 +1191,49 @@ struct CommGather : public MultiPlanItem
     std::vector<hipEvent_wrapper_t> events;
 };
 
+// Send data from all ranks to all ranks in the plan.  Each rank must
+// send from/to a single buffer (with different read/write offsets
+// for each other rank).
+struct CommAllToAllv : public MultiPlanItem
+{
+    CommAllToAllv() = default;
+
+    rocfft_precision  precision;
+    rocfft_array_type arrayType;
+
+    // counts and offsets are all in elements (where element size is
+    // knowable from precision + array type), for the current rank
+    std::vector<size_t> sendOffsets;
+    std::vector<size_t> sendCounts;
+    std::vector<size_t> recvOffsets;
+    std::vector<size_t> recvCounts;
+
+    // send/receive buffers
+    BufferPtr sendBuf;
+    BufferPtr recvBuf;
+
+    void ExecuteAsync(const rocfft_plan     plan,
+                      void*                 in_buffer[],
+                      void*                 out_buffer[],
+                      rocfft_execution_info info,
+                      size_t                multiPlanIdx) override;
+    void Wait() override;
+
+    void Print(rocfft_ostream& os, const int indent) const override;
+
+    bool WritesToBuffer(const BufferPtr& ptr) const override
+    {
+        // only writes to receive buffer
+        return ptr == recvBuf;
+    }
+
+    bool ExecutesOnRank(int comm_rank) const override
+    {
+        // runs on all ranks
+        return true;
+    }
+};
+
 // Tree-structured FFT plan.  This is specific to a single device on
 // a single rank, since the TreeNodes inside here will have device
 // memory allocated for things like kernel arguments and twiddles.
