@@ -31,24 +31,17 @@
 #include <string>
 #include <vector>
 
+#include "client_except.h"
 #include "enum_to_string.h"
 #include "fft_params.h"
 #include "fftw_transform.h"
 #include "gpubuf.h"
-#include "gtest_except.h"
 #include "rocfft_against_fftw.h"
+#include "sys_mem.h"
 #include "test_params.h"
 
-extern int    verbose;
-extern size_t ramgb;
-extern bool   fftw_compare;
-
-static const size_t ONE_GiB = 1 << 30;
-
-inline size_t bytes_to_GiB(const size_t bytes)
-{
-    return bytes == 0 ? 0 : (bytes - 1 + ONE_GiB) / ONE_GiB;
-}
+extern int  verbose;
+extern bool fftw_compare;
 
 // Remember the results of the last FFT we computed with FFTW.  Tests
 // are ordered so that later cases can often reuse this result.
@@ -67,79 +60,6 @@ struct last_cpu_fft_cache
 };
 extern last_cpu_fft_cache last_cpu_fft_data;
 
-struct system_memory
-{
-    size_t total_bytes = 0;
-    size_t free_bytes  = 0;
-    // Limits the amount of memory used throught the tests.
-    // Test RAM limit is: total_bytes * percentage_usable_memory.
-    // If the test system has little swap space (or there are many
-    // other processes running concurrently) and we are too
-    // aggressive with host memory usage, then the test process
-    // may get OOM killed.
-    static constexpr double percentage_usable_memory = .85;
-};
-extern system_memory start_memory;
-
-system_memory get_system_memory();
-
-// Estimate the amount of host memory needed for buffers.
-inline size_t needed_ram_buffers(const fft_params& params, const int verbose)
-{
-    // This calculation is assuming contiguous data but noncontiguous buffers
-    // are assumed to require a close enough amount of space for the purposes
-    // of this estimate.
-
-    size_t needed_ram = 6
-                        * std::accumulate(params.length.begin(),
-                                          params.length.end(),
-                                          static_cast<size_t>(1),
-                                          std::multiplies<size_t>());
-
-    // Account for precision and data type:
-    if(params.transform_type != fft_transform_type_real_forward
-       && params.transform_type != fft_transform_type_real_inverse)
-    {
-        needed_ram *= 2;
-    }
-    switch(params.precision)
-    {
-    case fft_precision_half:
-        needed_ram *= 2;
-        break;
-    case fft_precision_single:
-        needed_ram *= 4;
-        break;
-    case fft_precision_double:
-        needed_ram *= 8;
-        break;
-    }
-
-    needed_ram *= params.nbatch;
-
-    if(verbose)
-    {
-        std::cout << "required host memory for buffers (GiB): " << bytes_to_GiB(needed_ram) << "\n";
-    }
-
-    return needed_ram;
-}
-
-// Check if buffers will fit in host memory. Returns the amount of ram needed.
-template <class Tparams>
-inline void
-    check_problem_fits_host_memory(const Tparams& params, const int verbose, size_t& needed_ram)
-{
-    needed_ram = needed_ram_buffers(params, verbose);
-
-    if(ramgb > 0 && needed_ram > ramgb * ONE_GiB)
-    {
-        std::stringstream msg;
-        msg << "needed_ramgb: " << bytes_to_GiB(needed_ram) << ", ramgb limit: " << ramgb << ".\n";
-        throw ROCFFT_GTEST_SKIP{std::move(msg)};
-    }
-}
-
 // Perform several checks to make sure buffers will fit in device memory
 template <class Tparams>
 inline void check_problem_fits_device_memory(Tparams& params, const int verbose)
@@ -163,11 +83,11 @@ inline void check_problem_fits_device_memory(Tparams& params, const int verbose)
                 ss << "hipMemGetInfo failure with error " << hip_status;
             if(skip_runtime_fails)
             {
-                throw ROCFFT_GTEST_SKIP{std::move(ss)};
+                throw ROCFFT_SKIP{ss.str()};
             }
             else
             {
-                throw ROCFFT_GTEST_FAIL{std::move(ss)};
+                throw ROCFFT_FAIL{ss.str()};
             }
         }
         vram_avail = total;
@@ -187,7 +107,7 @@ inline void check_problem_fits_device_memory(Tparams& params, const int verbose)
         std::stringstream ss;
         ss << "Raw problem size (" << bytes_to_GiB(raw_vram_footprint)
            << " GiB) raw data too large for device";
-        throw ROCFFT_GTEST_SKIP{std::move(ss)};
+        throw ROCFFT_SKIP{ss.str()};
     }
 
     if(verbose > 2)
@@ -208,7 +128,7 @@ inline void check_problem_fits_device_memory(Tparams& params, const int verbose)
         std::stringstream ss;
         ss << "Problem size (" << bytes_to_GiB(vram_footprint)
            << " GiB) raw data too large for device";
-        throw ROCFFT_GTEST_SKIP{std::move(ss)};
+        throw ROCFFT_SKIP{ss.str()};
     }
 }
 
@@ -226,72 +146,6 @@ bool fftw_plan_uses_bluestein(const typename fftw_trait<Tfloat>::fftw_plan_type&
 #endif
 }
 
-// Estimate the amount of host memory needed for fftw.
-template <typename Tfloat>
-inline size_t needed_ram_fftw(const fft_params&                                  contiguous_params,
-                              const typename fftw_trait<Tfloat>::fftw_plan_type& cpu_plan,
-                              const int                                          verbose)
-{
-    size_t total_length = std::accumulate(contiguous_params.length.begin(),
-                                          contiguous_params.length.end(),
-                                          static_cast<size_t>(1),
-                                          std::multiplies<size_t>());
-    size_t needed_ram   = 0;
-    // Detect Bluestein in plan
-    if(fftw_plan_uses_bluestein<Tfloat>(cpu_plan))
-    {
-        for(size_t dim : contiguous_params.length)
-        {
-            unsigned int needed_ram_dim = dim;
-
-            // Next-plus-one-power-of-two multiplied any other lengths
-            needed_ram_dim--;
-
-            needed_ram_dim |= needed_ram_dim >> 2;
-            needed_ram_dim |= needed_ram_dim >> 4;
-            needed_ram_dim |= needed_ram_dim >> 8;
-            needed_ram_dim |= needed_ram_dim >> 16;
-
-            needed_ram_dim++;
-
-            needed_ram_dim *= 2 * (total_length / dim);
-
-            if(needed_ram_dim > needed_ram)
-            {
-                needed_ram = needed_ram_dim;
-            }
-        }
-    }
-
-    // Account for precision and data type:
-    if(contiguous_params.transform_type != fft_transform_type_real_forward
-       && contiguous_params.transform_type != fft_transform_type_real_inverse)
-    {
-        needed_ram *= 2;
-    }
-    switch(contiguous_params.precision)
-    {
-    case fft_precision_half:
-        needed_ram *= 2;
-        break;
-    case fft_precision_single:
-        needed_ram *= 4;
-        break;
-    case fft_precision_double:
-        needed_ram *= 8;
-        break;
-    }
-
-    needed_ram *= contiguous_params.nbatch;
-
-    if(verbose)
-    {
-        std::cout << "required host memory for FFTW (GiB): " << bytes_to_GiB(needed_ram) << "\n";
-    }
-
-    return needed_ram;
-}
-
 // Base gtest class for comparison with FFTW.
 class accuracy_test : public ::testing::TestWithParam<fft_params>
 {
@@ -407,13 +261,15 @@ inline void execute_gpu_fft(Tparams&              params,
         if(hip_status != hipSuccess)
         {
             ++n_hip_failures;
+            std::stringstream ss;
+            ss << "Error occurred when allocating device memory for loading callback";
             if(skip_runtime_fails)
             {
-                throw ROCFFT_GTEST_SKIP();
+                throw ROCFFT_SKIP{ss.str()};
             }
             else
             {
-                throw ROCFFT_GTEST_FAIL();
+                throw ROCFFT_FAIL{ss.str()};
             }
         }
         hip_status = hipMemcpy(load_cb_data_dev.data(),
@@ -423,13 +279,15 @@ inline void execute_gpu_fft(Tparams&              params,
         if(hip_status != hipSuccess)
         {
             ++n_hip_failures;
+            std::stringstream ss;
+            ss << "Error occurred when copying data to device for loading callback";
             if(skip_runtime_fails)
             {
-                throw ROCFFT_GTEST_SKIP();
+                throw ROCFFT_SKIP{ss.str()};
             }
             else
             {
-                throw ROCFFT_GTEST_FAIL();
+                throw ROCFFT_FAIL{ss.str()};
             }
         }
 
@@ -453,13 +311,15 @@ inline void execute_gpu_fft(Tparams&              params,
         if(hip_status != hipSuccess)
         {
             ++n_hip_failures;
+            std::stringstream ss;
+            ss << "Error occurred when allocating device memory for storing callback";
             if(skip_runtime_fails)
             {
-                throw ROCFFT_GTEST_SKIP();
+                throw ROCFFT_SKIP{ss.str()};
             }
             else
             {
-                throw ROCFFT_GTEST_FAIL();
+                throw ROCFFT_FAIL{ss.str()};
             }
         }
 
@@ -470,13 +330,15 @@ inline void execute_gpu_fft(Tparams&              params,
         if(hip_status != hipSuccess)
         {
             ++n_hip_failures;
+            std::stringstream ss;
+            ss << "Error occurred when copying data to device for storing callback";
             if(skip_runtime_fails)
             {
-                throw ROCFFT_GTEST_SKIP();
+                throw ROCFFT_SKIP{ss.str()};
             }
             else
             {
-                throw ROCFFT_GTEST_FAIL();
+                throw ROCFFT_FAIL{ss.str()};
             }
         }
 
@@ -519,11 +381,11 @@ inline void execute_gpu_fft(Tparams&              params,
             ss << "hipMemcpy failure";
             if(skip_runtime_fails)
             {
-                throw ROCFFT_GTEST_SKIP{std::move(ss)};
+                throw ROCFFT_SKIP{ss.str()};
             }
             else
             {
-                throw ROCFFT_GTEST_FAIL{std::move(ss)};
+                throw ROCFFT_FAIL{ss.str()};
             }
         }
     }
@@ -708,11 +570,11 @@ inline void run_round_trip_inverse(Tparams&              params,
         ++n_hip_failures;
         if(skip_runtime_fails)
         {
-            throw ROCFFT_GTEST_SKIP{std::move(ss)};
+            throw ROCFFT_SKIP{ss.str()};
         }
         else
         {
-            throw ROCFFT_GTEST_FAIL{std::move(ss)};
+            throw ROCFFT_FAIL{ss.str()};
         }
     }
     ASSERT_EQ(plan_status, fft_status_success) << "round trip inverse plan creation failed";
@@ -738,11 +600,11 @@ inline void run_round_trip_inverse(Tparams&              params,
                     ss << "hipMemset failure";
                     if(skip_runtime_fails)
                     {
-                        throw ROCFFT_GTEST_SKIP{std::move(ss)};
+                        throw ROCFFT_SKIP{ss.str()};
                     }
                     else
                     {
-                        throw ROCFFT_GTEST_FAIL{std::move(ss)};
+                        throw ROCFFT_FAIL{ss.str()};
                     }
                 }
             }
@@ -891,9 +753,7 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
     auto ibuffer_sizes = params.ibuffer_sizes();
     auto obuffer_sizes = params.obuffer_sizes();
 
-    // Make sure FFT buffers fit in host/device memory
-    size_t needed_ram;
-    check_problem_fits_host_memory(params, verbose, needed_ram);
+    // Make sure FFT buffers fit in device memory
     check_problem_fits_device_memory(params, verbose);
 
     // Create FFT plan - this will also allocate work buffer, but
@@ -910,11 +770,11 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
         ss << "Work buffer allocation failed with size: " << params.workbuffersize;
         if(skip_runtime_fails)
         {
-            throw ROCFFT_GTEST_SKIP{std::move(ss)};
+            throw ROCFFT_SKIP{ss.str()};
         }
         else
         {
-            throw ROCFFT_GTEST_FAIL{std::move(ss)};
+            throw ROCFFT_FAIL{ss.str()};
         }
     }
     ASSERT_EQ(plan_status, fft_status_success) << "plan creation failed";
@@ -949,17 +809,25 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
         if(hip_status != hipSuccess)
         {
             std::stringstream ss;
-            ss << "hipMalloc failure for input buffer " << i << " size " << ibuffer_sizes[i] << "("
-               << bytes_to_GiB(ibuffer_sizes[i]) << " GiB)"
-               << " with code " << hipError_to_string(hip_status);
-            ++n_hip_failures;
-            if(skip_runtime_fails)
+            if(hip_status == hipErrorOutOfMemory)
             {
-                throw ROCFFT_GTEST_SKIP{std::move(ss)};
+                ss << "Input buffer size (" << bytes_to_GiB(ibuffer_sizes[i])
+                   << " GiB) raw data too large for device";
             }
             else
             {
-                throw ROCFFT_GTEST_FAIL{std::move(ss)};
+                ss << "hipMalloc failure for input buffer " << i << " size " << ibuffer_sizes[i]
+                   << "(" << bytes_to_GiB(ibuffer_sizes[i]) << " GiB)"
+                   << " with code " << hipError_to_string(hip_status);
+            }
+            ++n_hip_failures;
+            if(skip_runtime_fails)
+            {
+                throw ROCFFT_SKIP{ss.str()};
+            }
+            else
+            {
+                throw ROCFFT_FAIL{ss.str()};
             }
         }
         pibuffer[i] = ibuffer[i].data();
@@ -1024,19 +892,19 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
                     if(last_cpu_fft_data.precision == fft_precision_double)
                     {
                         convert_cpu_output_precision = std::async(std::launch::async, [&]() {
-                            narrow_precision_inplace<double, _Float16>(cpu_output.front());
+                            narrow_precision_inplace<double, rocfft_fp16>(cpu_output.front());
                         });
                         convert_cpu_input_precision  = std::async(std::launch::async, [&]() {
-                            narrow_precision_inplace<double, _Float16>(cpu_input.front());
+                            narrow_precision_inplace<double, rocfft_fp16>(cpu_input.front());
                         });
                     }
                     else if(last_cpu_fft_data.precision == fft_precision_single)
                     {
                         convert_cpu_output_precision = std::async(std::launch::async, [&]() {
-                            narrow_precision_inplace<float, _Float16>(cpu_output.front());
+                            narrow_precision_inplace<float, rocfft_fp16>(cpu_output.front());
                         });
                         convert_cpu_input_precision  = std::async(std::launch::async, [&]() {
-                            narrow_precision_inplace<float, _Float16>(cpu_input.front());
+                            narrow_precision_inplace<float, rocfft_fp16>(cpu_input.front());
                         });
                     }
                     else
@@ -1094,95 +962,154 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
                                                 contiguous_params.transform_type,
                                                 cpu_input,
                                                 cpu_output);
-
-        needed_ram += needed_ram_fftw<Tfloat>(contiguous_params, cpu_plan, verbose);
-
-        if(ramgb > 0 && needed_ram > ramgb * ONE_GiB)
-        {
-            if(verbose)
-            {
-                std::cout << "Problem exceeds memory limit; skipped [rocfft_transform]."
-                          << std::endl;
-            }
-            throw ROCFFT_GTEST_SKIP();
-        }
     }
 
+    // Host-side buffer used to store input to transfer to GPU, copy GPU input
+    // to contiguous layout for FFTW, and reused to store IFFT output.
     std::vector<hostbuf> gpu_input_data;
 
+    auto is_host_gen = (params.igen == fft_input_generator_host
+                        || params.igen == fft_input_random_generator_host);
+
     // allocate and populate the input buffer (cpu/gpu)
     if(run_fftw)
     {
         gpu_input_data = allocate_host_buffer(params.precision, params.itype, ibuffer_sizes_elems);
 
-        //generate the input directly on the gpu
-        params.compute_input(ibuffer);
-
-        // Copy the input to CPU
-        if(params.itype != contiguous_params.itype || params.istride != contiguous_params.istride
-           || params.idist != contiguous_params.idist || params.isize != contiguous_params.isize)
+#ifdef USE_HIPRAND
+        if(!is_host_gen)
         {
-            // Copy input to CPU
-            for(unsigned int idx = 0; idx < ibuffer.size(); ++idx)
+            // generate the input directly on the gpu
+            params.compute_input(ibuffer);
+
+            // Copy the input to CPU
+            if(params.itype != contiguous_params.itype
+               || params.istride != contiguous_params.istride
+               || params.idist != contiguous_params.idist
+               || params.isize != contiguous_params.isize)
             {
-                hip_status = hipMemcpy(gpu_input_data.at(idx).data(),
-                                       ibuffer[idx].data(),
-                                       ibuffer_sizes[idx],
-                                       hipMemcpyDeviceToHost);
-                if(hip_status != hipSuccess)
+                // Copy input to CPU
+                for(unsigned int idx = 0; idx < ibuffer.size(); ++idx)
                 {
-                    ++n_hip_failures;
-                    std::stringstream ss;
-                    ss << "hipMemcpy failure with error " << hip_status;
-                    if(skip_runtime_fails)
+                    hip_status = hipMemcpy(gpu_input_data.at(idx).data(),
+                                           ibuffer[idx].data(),
+                                           ibuffer_sizes[idx],
+                                           hipMemcpyDeviceToHost);
+                    if(hip_status != hipSuccess)
                     {
-                        throw ROCFFT_GTEST_SKIP{std::move(ss)};
+                        ++n_hip_failures;
+                        std::stringstream ss;
+                        ss << "hipMemcpy failure with error " << hip_status;
+                        if(skip_runtime_fails)
+                        {
+                            throw ROCFFT_SKIP{ss.str()};
+                        }
+                        else
+                        {
+                            throw ROCFFT_FAIL{ss.str()};
+                        }
                     }
-                    else
+                }
+
+                copy_buffers(gpu_input_data,
+                             cpu_input,
+                             params.ilength(),
+                             params.nbatch,
+                             params.precision,
+                             params.itype,
+                             params.istride,
+                             params.idist,
+                             contiguous_params.itype,
+                             contiguous_params.istride,
+                             contiguous_params.idist,
+                             params.ioffset,
+                             contiguous_params.ioffset);
+            }
+            else
+            {
+                // Copy input to CPU
+                for(unsigned int idx = 0; idx < ibuffer.size(); ++idx)
+                {
+                    hip_status = hipMemcpy(cpu_input.at(idx).data(),
+                                           ibuffer[idx].data(),
+                                           ibuffer_sizes[idx],
+                                           hipMemcpyDeviceToHost);
+                    if(hip_status != hipSuccess)
                     {
-                        throw ROCFFT_GTEST_FAIL{std::move(ss)};
+                        ++n_hip_failures;
+                        std::stringstream ss;
+                        ss << "hipMemcpy failure with error " << hip_status;
+                        if(skip_runtime_fails)
+                        {
+                            throw ROCFFT_SKIP{ss.str()};
+                        }
+                        else
+                        {
+                            throw ROCFFT_FAIL{ss.str()};
+                        }
                     }
                 }
             }
-
-            copy_buffers(gpu_input_data,
-                         cpu_input,
-                         params.ilength(),
-                         params.nbatch,
-                         params.precision,
-                         params.itype,
-                         params.istride,
-                         params.idist,
-                         contiguous_params.itype,
-                         contiguous_params.istride,
-                         contiguous_params.idist,
-                         params.ioffset,
-                         contiguous_params.ioffset);
         }
-        else
+
+#endif
+        if(is_host_gen)
         {
-            // Copy input to CPU
-            for(unsigned int idx = 0; idx < ibuffer.size(); ++idx)
+            params.compute_input(gpu_input_data);
+            // Copy the input to CPU
+            if(params.itype != contiguous_params.itype
+               || params.istride != contiguous_params.istride
+               || params.idist != contiguous_params.idist
+               || params.isize != contiguous_params.isize)
             {
-                hip_status = hipMemcpy(cpu_input.at(idx).data(),
-                                       ibuffer[idx].data(),
-                                       ibuffer_sizes[idx],
-                                       hipMemcpyDeviceToHost);
-                if(hip_status != hipSuccess)
+                // Copy input to CPU and make input contiguous
+                copy_buffers(gpu_input_data,
+                             cpu_input,
+                             params.ilength(),
+                             params.nbatch,
+                             params.precision,
+                             params.itype,
+                             params.istride,
+                             params.idist,
+                             contiguous_params.itype,
+                             contiguous_params.istride,
+                             contiguous_params.idist,
+                             params.ioffset,
+                             contiguous_params.ioffset);
+            }
+            else
+            {
+                // Direct copy input to cpu_input as is
+                for(unsigned int idx = 0; idx < gpu_input_data.size(); ++idx)
                 {
-                    ++n_hip_failures;
-                    std::stringstream ss;
-                    ss << "hipMemcpy failure with error " << hip_status;
-                    if(skip_runtime_fails)
+                    hip_status = hipMemcpy(cpu_input.at(idx).data(),
+                                           gpu_input_data.at(idx).data(),
+                                           gpu_input_data.at(idx).size(),
+                                           hipMemcpyHostToHost);
+                    if(hip_status != hipSuccess)
                     {
-                        throw ROCFFT_GTEST_SKIP{std::move(ss)};
-                    }
-                    else
-                    {
-                        throw ROCFFT_GTEST_FAIL{std::move(ss)};
+                        ++n_hip_failures;
+                        std::stringstream ss;
+                        ss << "hipMemcpy failure with error " << hip_status;
+                        if(skip_runtime_fails)
+                        {
+                            throw ROCFFT_SKIP{ss.str()};
+                        }
+                        else
+                        {
+                            throw ROCFFT_FAIL{ss.str()};
+                        }
                     }
                 }
             }
+            // Copy input to GPU
+            for(unsigned int idx = 0; idx < gpu_input_data.size(); ++idx)
+            {
+                hip_status = hipMemcpy(ibuffer[idx].data(),
+                                       gpu_input_data.at(idx).data(),
+                                       ibuffer_sizes[idx],
+                                       hipMemcpyHostToDevice);
+            }
         }
     }
     else if(fftw_compare)
@@ -1230,11 +1157,11 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
                 ss << "hipMemcpy failure with error " << hip_status;
                 if(skip_runtime_fails)
                 {
-                    throw ROCFFT_GTEST_SKIP{std::move(ss)};
+                    throw ROCFFT_SKIP{ss.str()};
                 }
                 else
                 {
-                    throw ROCFFT_GTEST_FAIL{std::move(ss)};
+                    throw ROCFFT_FAIL{ss.str()};
                 }
             }
         }
@@ -1296,11 +1223,11 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
                    << " with code " << hipError_to_string(hip_status);
                 if(skip_runtime_fails)
                 {
-                    throw ROCFFT_GTEST_SKIP{std::move(ss)};
+                    throw ROCFFT_SKIP{ss.str()};
                 }
                 else
                 {
-                    throw ROCFFT_GTEST_FAIL{std::move(ss)};
+                    throw ROCFFT_FAIL{ss.str()};
                 }
             }
 
@@ -1319,11 +1246,11 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
                     ss << "hipMemset failure with error " << hip_status;
                     if(skip_runtime_fails)
                     {
-                        throw ROCFFT_GTEST_SKIP{std::move(ss)};
+                        throw ROCFFT_SKIP{ss.str()};
                     }
                     else
                     {
-                        throw ROCFFT_GTEST_FAIL{std::move(ss)};
+                        throw ROCFFT_FAIL{ss.str()};
                     }
                 }
             }
@@ -1408,10 +1335,7 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
     // Compute the l-infinity and l-2 distance between the CPU and GPU output:
     // wait for cpu FFT so we can compute cutoff
 
-    const auto total_length = std::accumulate(params.length.begin(),
-                                              params.length.end(),
-                                              static_cast<size_t>(1),
-                                              std::multiplies<size_t>());
+    const auto total_length = product(params.length.begin(), params.length.end());
 
     std::unique_ptr<std::vector<std::pair<size_t, size_t>>> linf_failures;
     if(verbose > 1)
@@ -1541,7 +1465,7 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
             << "\tcutoff: " << linf_cutoff << params.str();
 
         EXPECT_TRUE(diff.l_2 / cpu_output_norm.l_2
-                    < sqrt(log2(total_length)) * type_epsilon(params.precision))
+                    <= sqrt(log2(total_length)) * type_epsilon(params.precision))
             << "L2 test failed. L2: " << diff.l_2
             << "\tnormalized L2: " << diff.l_2 / cpu_output_norm.l_2
             << "\tepsilon: " << sqrt(log2(total_length)) * type_epsilon(params.precision)
